from typing import Optional, Iterable, Type

from w2re.prefix_tree.primitives import PrefixTreeNode
from w2re.formatters import BaseFormatter


class PrefixTree:
    def __init__(self, words: Optional[Iterable[str]] = None) -> None:
        self._root_node = PrefixTreeNode()

        if words is not None:
            self.extend(words)

    def add(self, word: str):
        self._root_node.add(word)

    def extend(self, words: Iterable[str]):
        for word in words:
            self.add(word)

    def to_regexp(self, formatter: Type[BaseFormatter]) -> str:
        """
        :return Returns regular expression representation of the structure.
        If the structure is empty, returns regular expression matching
        empty string.
        """
        return formatter.wrap_regexp(self._root_node)



import re

from w2re.prefix_tree.letter_range_utils import collapse_letter_ranges


def get_stop_index(word, sliding_window_length):
    """ Stop index for two adjacent sliding windows.

    For blActgActgA it would be::

        blAct | gActg | A -> b | lActg | ActgA

    """
    return len(word) - sliding_window_length * 2 + 1


def add_brackets_around_string(word: str):
    """Adds square brackets for single character strings and round the rest.
    """
    if len(word) == 1 or (len(word) == 2 and word.startswith('\\')):
        return word

    return '(?:' + word + ')'


def compress(word, sliding_window_len=None):
    """
    This function uses sliding window of decreasing size to find repeated,
    non-overlapping and adjacent sub-strings.

    Example::

        BLACTGACTGA contains BL - 2x ACTG - A or BLA - 2x CTGA

    Note that the sub-string can be repeated more than 2x, so the function remembers
    first window and then next window, that can be repeated 1+ times, while
    counting the repetitions.

    The function also applies itself on the longest discovered sub-string
    recursively. So it discovers shorter sub-strings in sub-strings.

    :param word: Input string.
    :param sliding_window_len: Initial window size. Used by recursion. Do not set this
        yourself.
    :return: Compressed string.
    """
    if sliding_window_len is None:
        sliding_window_len = len(word) // 2

    if sliding_window_len <= 0 or len(word) <= 1:
        return word  # don't compress these

    stop_index = get_stop_index(word, sliding_window_len)
    i = 0
    compression_observed = False

    while i < stop_index:
        next_window_start = i + sliding_window_len  # there can be more than 2 windows
        first_window = word[i:next_window_start]
        repeat_count = 0

        while True:  # Repeat next windows until non-matching found
            next_window_end = next_window_start + sliding_window_len

            if next_window_end >= len(word) + 1:
                break  # next window overflows the string -> stop

            next_window = word[next_window_start:next_window_end]

            if first_window != next_window:
                break

            repeat_count += 1
            next_window_start = next_window_end

        if repeat_count > 0:
            compression_observed = True
            prefix = word[:i]

            repeated_block_len_end = i + sliding_window_len * (repeat_count + 1)
            suffix = word[repeated_block_len_end:]

            center = add_brackets_around_string(compress(first_window))

            new_word = prefix + center + '{' + str(repeat_count + 1) + '}'
            i = len(new_word) - 1  # skip the compressed part, investigate only the rest
            new_word += suffix
            word = new_word
            stop_index = get_stop_index(word, sliding_window_len)

        i += 1

    return word if compression_observed else compress(word, sliding_window_len - 1)


class PrefixTreeNode:
    def __init__(self, terminal_node=False):
        self._edges = None
        self.terminal_node = terminal_node

    def add(self, word: str, new_node: "PrefixTreeNode" = None):
        if not word:
            return

        first_letter = word[0]

        # no edge starting with the same letter, create a new branch
        if self._edges is None or first_letter not in self._edges:
            if self._edges is None:
                self._edges = {}

            self._edges[first_letter] = PrefixTreeEdge(word, True, new_node)
        else:
            self._edges[first_letter].add(word)

    @staticmethod
    def _non_matching_brackets(string: str) -> str:
        return '(?:' + string + ')'

    def _add_me(self, string: str, add_brackets: bool) -> str:
        if self.terminal_node:
            return (self._non_matching_brackets(string) if add_brackets else string) + '?'

        return string

    def to_regexp(self):
        if self._edges is None:  # target node is leaf
            return ''

        # not leaf, get sub-strings from children
        letters = []
        strings = []

        for edge in self._edges.values():
            sub_str = edge.to_regexp()
            if len(sub_str) == 1:  # just one letter received, more can be grouped in [XYZ]
                letters.append(sub_str)
            else:
                # strings received, can be grouped just with | symbol
                strings.append(sub_str)

        if letters and not strings:  # just letters
            if len(letters) == 1:  # just one letter
                return self._add_me(re.escape(letters[0]), False)

            # more letters, group in []
            return self._add_me(
                '[' + ''.join(collapse_letter_ranges(letters)) + ']', False
            )

        if not letters and len(strings) == 1:  # just one string
            return self._add_me(strings[0], True)

        # combination of letters and strings
        if len(letters) > 1:
            strings.insert(0, '[' + ''.join(collapse_letter_ranges(letters)) + ']')
        elif len(letters) == 1:
            strings.insert(0, letters[0])

        return self._add_me(self._non_matching_brackets('|'.join(strings)), False)


class PrefixTreeEdge:
    def __init__(self, label, terminal, new_node=None):
        self._target_node = new_node if new_node is not None else PrefixTreeNode(terminal)
        self._label = label

    def _split(self, word, position):
        new_node = PrefixTreeNode(True)
        new_node.add(self._label[position:], self._target_node)
        self._label = word
        self._target_node = new_node

    def _branch(self, word, position):
        new_node = PrefixTreeNode(False)  # create branching node, not terminal

        # create branch from the original label, use original node as end of the edge
        new_node.add(self._label[position:], self._target_node)

        # create branch from the new word, new node will be needed
        new_node.add(word[position:])
        self._label = word[:position]  # update label to the common part of both
        self._target_node = new_node  # update target node to point to the splitting one

    def add(self, word):
        # find branching
        for i in range(min(len(word), len(self._label))):
            if word[i] != self._label[i]:  # difference found, branch
                self._branch(word, i)
                return

        # no differences on the shortest length of both label and the new word
        if len(word) > len(self._label):
            # current edge is prefix of inputted word
            self._target_node.add(word[len(self._label):])
        elif len(word) < len(self._label):
            # current edge has inputted word as prefix, split it
            self._split(word, len(word))
        else:
            # current word and edge are the same, mark target node as terminal
            self._target_node.terminal_node = True

    def to_regexp(self):
        from_below = self._target_node.to_regexp()

        if len(from_below) + len(self._label) <= 1:  # don't escape single characters
            return self._label + from_below

        return compress(re.escape(self._label)) + from_below



#!/usr/bin/python
import argparse
import sys
from typing import (  # pylint: disable=unused-import; false positive
    Dict,
    Type,
)

from w2re import (
    APPLICATION_NAME,
    CHANGELOG_URL,
    __version__ as VERSION
)
from w2re.formatters import (  # pylint: disable=unused-import; false positive
    ALL_FORMATTERS,
    BaseFormatter,
)
from w2re.utils import stream_to_regexp

FORMATTERS_BY_CODE = {
    formatter.code(): formatter
    for formatter in ALL_FORMATTERS
}  # type: Dict[str, Type[BaseFormatter]]


APPLICATION_DESCRIPTION = (
    APPLICATION_NAME + ': Words to Regular Expression, script for converting '
    'a list of words into compressed regular expression. Empty lines are ignored.'
)


def main(mock_args=None):
    parser = argparse.ArgumentParser(
        description=APPLICATION_DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter
    )

    formats_list = '\n'.join(
        "%s:\t%s" % (key, formatter.description())
        for key, formatter in FORMATTERS_BY_CODE.items()
    )

    default_formatter_code = ALL_FORMATTERS[0].code()

    parser.add_argument(
        '-f',
        dest='formatter',
        default=default_formatter_code,
        metavar='<type>',
        choices=tuple(FORMATTERS_BY_CODE.keys()),
        help="Output format. Default is '{}'. Possible value are:\n\n{}.".format(
            default_formatter_code, formats_list
        )
    )

    parser.add_argument(
        '-i',
        dest='input',
        default=sys.stdin,
        metavar='<filename>',
        type=argparse.FileType('r'),
        help='Input file. If none specified, stdin will be used instead.'
    )

    parser.add_argument(
        '--version',
        dest='show_version',
        default=False,
        action='store_true',
        help='Show version information.'
    )

    args = parser.parse_args(mock_args)

    if args.show_version:
        print('{} {}\n\nFor changelog, see: {}'.format(
            APPLICATION_NAME, VERSION, CHANGELOG_URL
        ))
    else:
        print(stream_to_regexp(args.input, FORMATTERS_BY_CODE[args.formatter]), end='')


if __name__ == '__main__':  # pragma: no cover
    main()



from typing import (
    List,
)

_SQUARE_BRACKET_ESCAPABLES = {'[', ']', '\\'}


def escape_char_in_square_brackets(character: str) -> str:
    return '\\' + character if character in _SQUARE_BRACKET_ESCAPABLES else character


def _formatted_letter_range(first_letter: str, last_letter: str) -> List[str]:
    if first_letter == last_letter:
        raise RuntimeError(
            "first_letter and last_letter must not be the same."
        )  # pragma: no cover

    if ord(first_letter) + 1 < ord(last_letter):
        return [first_letter + '-' + last_letter]

    return [first_letter, last_letter]


def _is_next_letter(first_letter: str, second_letter: str) -> bool:
    return ord(first_letter) + 1 == ord(second_letter)


_STATE_NO_RANGE = 1
_STATE_RANGE_START = 2
_STATE_IN_RANGE = 3


# TODO(Radek): Implement proper FSM
def collapse_letter_ranges(  # pylint: disable=too-many-branches; required for FSM implementation
        letters: List[str]
) -> List[str]:
    if len(letters) == 1:
        return [escape_char_in_square_brackets(letters[0])]

    # Make hyphen first in the list, if present
    letters.sort(key=lambda character: -1 if character == '-' else ord(character))
    letters_out = []  # type: List[str]
    state = _STATE_NO_RANGE
    first_letter = ''
    previous_letter = ''
    last_index = len(letters)

    for index, current_letter in enumerate(letters, 1):
        if state == _STATE_NO_RANGE:
            if current_letter.isalnum():
                state = _STATE_RANGE_START
                first_letter = current_letter
            else:
                letters_out.append(escape_char_in_square_brackets(current_letter))

        elif state == _STATE_RANGE_START:
            if not current_letter.isalnum():
                state = _STATE_NO_RANGE
                letters_out.append(first_letter)
                letters_out.append(escape_char_in_square_brackets(current_letter))
            elif _is_next_letter(previous_letter, current_letter):
                state = _STATE_IN_RANGE
            else:
                letters_out.append(first_letter)
                first_letter = current_letter

        elif state == _STATE_IN_RANGE:
            if not _is_next_letter(previous_letter, current_letter):
                state = _STATE_RANGE_START
                letters_out.extend(_formatted_letter_range(first_letter, previous_letter))
                first_letter = current_letter
        else:
            raise RuntimeError("Unexpected state: " + str(state))  # pragma: no cover

        if index == last_index:
            if state == _STATE_RANGE_START:
                letters_out.append(current_letter)
            elif state == _STATE_IN_RANGE:
                letters_out.extend(_formatted_letter_range(first_letter, current_letter))

        previous_letter = current_letter

    return letters_out



from w2re.formatters import (
    BaseFormatter,
    PythonFormatter,
    PythonWordMatchFormatter,
)
from w2re.prefix_tree.tree import PrefixTree
from w2re.utils import (
    iterable_to_regexp,
    stream_to_regexp,
)

__version__ = '3.1.0'

APPLICATION_NAME = 'w2re'
CHANGELOG_URL = 'https://github.com/radeklat/words-to-regular-expression/' \
                'blob/develop/CHANGELOG.md'



import os
from typing import (
    TextIO,
    Type,
    Iterable,
)

from w2re import PythonFormatter
from w2re.formatters import BaseFormatter
from w2re.prefix_tree.tree import PrefixTree


def stream_to_regexp(
        stream: TextIO,
        formatter: Type[BaseFormatter] = PythonFormatter
) -> str:
    lines_generator = (line for line in stream.read().split(os.linesep) if line)
    prefix_tree = PrefixTree(lines_generator)
    return prefix_tree.to_regexp(formatter)


def iterable_to_regexp(
        iterable: Iterable[str],
        formatter: Type[BaseFormatter] = PythonFormatter
) -> str:
    return PrefixTree(iterable).to_regexp(formatter)



from w2re.prefix_tree.primitives import PrefixTreeNode


class BaseFormatter:
    _DESCRIPTION = ''
    _CODE = ''

    @staticmethod
    def wrap_regexp(root_node: PrefixTreeNode) -> str:
        raise NotImplementedError()

    @classmethod
    def description(cls):
        assert cls._DESCRIPTION, "Sub-classes must override _DESCRIPTION"
        return cls._DESCRIPTION

    @classmethod
    def code(cls):
        assert cls._CODE, "Sub-classes must override _CODE"
        return cls._CODE


class PythonFormatter(BaseFormatter):
    _DESCRIPTION = 'Python regular expression'
    _CODE = 'py'
    _EMPTY_STRING_MATCH = r"\A\Z"

    @staticmethod
    def wrap_regexp(root_node: PrefixTreeNode) -> str:
        regexp = root_node.to_regexp()

        if regexp:
            return regexp

        return PythonFormatter._EMPTY_STRING_MATCH


class PythonWordMatchFormatter(PythonFormatter):
    _DESCRIPTION = 'Python word matching regular expression'
    _CODE = 'pyw'

    @staticmethod
    def wrap_regexp(root_node: PrefixTreeNode) -> str:
        regexp = root_node.to_regexp()

        if regexp:
            return r"(?:\W+|\A)({})(?=\W+|\Z)".format(regexp)

        return PythonFormatter._EMPTY_STRING_MATCH


ALL_FORMATTERS = (PythonFormatter, PythonWordMatchFormatter)
